/*
 * stream filter to change printable ascii from "btoa" back into 8 bit bytes
 * if bad chars, or Csums do not match: exit(1) [and NO output]
 *
 *  Paul Rutter         Joe Orost
 *  philabs!per         petsd!joe
 */
#ifdef  DOCUMENTATION

title   atob    Ascii to Binary Conversion
index           Ascii to binary conversion

synopsis

        atob source destination

description

        atob reads a file created by btoa, converting it to
        its original binary format.  The checksums calculated
        by btoa are checked.

        If both files are missing, atob reads from the standard
        input and writes to the standard output.  Because of
        RMS file attributes, this may not work well on VMS
        or RSX-11M.

warning

        Neither btoa nor atob understand RMS file attributes.
        The output of atob is a "binary" file if it was specified
        on the command line, and a text (or stream) file if
        redirection was used.

vms note

        Redirection is intentionally missing from the VMS version.

author

        Paul Rutter     philabs!per

        Joe Orost       petsd!joe

        Martin Minow    minow%thundr.dec@decwrl.dec.com

#endif

/*)BUILD
                $(CPP)          = 1
                $(TKBOPTIONS)   = {
                        TASK    = ...ATB
                }
*/

#include <stdio.h>

#ifdef  vms
#include                errno
#include                ssdef
#include                stsdef
#define IO_SUCCESS      (SS$_NORMAL | STS$M_INHIB_MSG)
#define IO_ERROR        (SS$_ABORT)
#define unlink(x)       delete(x)
#endif
/*
 * Define exit() codes.
 */
#ifndef IO_SUCCESS
#define IO_SUCCESS      0                       /* Normal exit          */
#define IO_ERROR        1                       /* Error exit           */
#endif

#define reg register

#ifndef decus
#define streq(s0, s1)  strcmp(s0, s1) == 0
#endif

long int Ceor = 0;
long int Csum = 0;
long int Crot = 0;
long int bcount = 0;
long int word = 0;
FILE *tempfd;


fatal() {
  fprintf(stderr, "bad format or Csum to atob\n");
  exit(IO_ERROR);
}


#define DE(c) ((c) - ' ')

decode(c) reg c;
{
  if (c == 'z') {
    if (bcount != 0) {
      fatal();
    }
    else {
      byteout(0);
      byteout(0);
      byteout(0);
      byteout(0);
    }
  }
  else if ((c >= ' ') && (c < (' ' + 85))) {
    if (bcount == 0) {
      word = DE(c);
      ++bcount;
    }
    else if (bcount < 4) {
      word *= 85;
      word += DE(c);
      ++bcount;
    }
    else {
      word = ((unsigned) word * (unsigned) 85) + DE(c);
      byteout((word >> 24) & 255);
      byteout((word >> 16) & 255);
      byteout((word >> 8) & 255);
      byteout(word & 255);
      word = 0;
      bcount = 0;
    }
  }
  else {
    fatal();
  }
}

byteout(c) reg c;
{
  Ceor ^= c;
  Csum += c;
  Csum += 1;
  if ((Crot & 0x80000000)) {
    Crot <<= 1;
    Crot += 1;
  }
  else {
    Crot <<= 1;
  }
  Crot += c;
  putc(c, tempfd);
}

main(argc, argv) char **argv;
{
  reg c;
  reg long int i;
  char tmpname[100];
  char buf[100];
  long int n1, n2, oeor, osum, orot;

  if (argc == 3) {
    if (freopen(argv[1], "r", stdin) == NULL) {
        perror(argv[1]);
        fatal();
    }
    if (freopen(argv[2], "wb", stdout) == NULL) {
        perror(argv[2]);
        fatal();
    }
  }
  else if (argc != 1) {
    fprintf(stderr,"bad args to %s\n", argv[0]);
    exit(IO_ERROR);
  }
#ifdef vms
  sprintf(tmpname, "%x.atob_temp;", getpid());          /* VMS          */
#else
#ifdef rsx
  strcpy(tmpname, "atob.tmp;");                         /* RSX-11M      */
#else
#ifdef rt11
  strcpy(tmpname, "atob.tmp");                          /* RT11         */
#else
  sprintf(tmpname, "/usr/tmp/atob.%x", getpid());       /* else Unix    */
#endif
#endif
#endif
#ifdef decus
  tempfd = fopen(tmpname, "wb");
#else
  tempfd = fopen(tmpname, "wb+");
#endif
  if (tempfd == NULL) {
    perror(tmpname);
    fatal();
  }
  /*search for header line*/
  for (;;) {
    if (fgets(buf, sizeof buf, stdin) == NULL) {
      fatal();
    }
    if (streq(buf, "xbtoa Begin\n")) {
      break;
    }
  }

  while ((c = getchar()) != EOF) {
    if (c == '\n') {
      continue;
    }
    else if (c == 'x') {
      break;
    }
    else {
      decode(c);

    }
  }
  if (scanf("btoa End N %d %x E %x S %x R %x\n", &n1, &n2, &oeor, &osum, &orot)
 != 5) {
    fatal();
  }
  if ((n1 != n2) || (oeor != Ceor) || (osum != Csum) || (orot != Crot)) {
    fatal();
  }
  else {
    /*copy OK tmp file to stdout*/;
#ifdef decus
    fclose(tempfd)
    if (tempfd = fopen(tmpname, "rb")) == NULL) {
      perror(tmpname);
      fatal();
    }
#else
    fseek(tempfd, 0, 0);
#endif
    for (i = n1; --i >= 0;) {
      putchar(getc(tempfd));
    }
    fclose(tempfd);             /* 89.07.10     */
    unlink(tmpname);
  }
  exit(IO_SUCCESS);
}
